import h5py
import numpy as np
import os

from uncertainties import ufloat

import unittest
from nose.tools import assert_raises, assert_almost_equal
from numpy.testing import assert_array_almost_equal

from brownian import (Cantilever, BrownianMotionFitter, u, calc_k_c,
                      translate_fit_parameters, avg_ci_data, get_data,
                      fit_residuals, convert_data, silentremove)

# Make sure we are executing from the current directory. See
# http://stackoverflow.com/q/9887259/2823213
os.chdir(os.path.dirname(os.path.abspath(__file__)))

# Load some test data from an hdf5 file.
# TODO: Get rid of dependence on .h5 file; this data can be programatically
#       generated.
with h5py.File('data.h5', 'r') as data:
    f = data['f'].value
    PSDx = data['PSDx'].value
    PSD_wgt = data['PSDw'].value

est_cant = Cantilever(f_c=63700*u.Hz, k_c=3.5*u.N/u.m,
                      Q=20000*u.dimensionless)
T = 295


class testBrownianMotionFitter_init(unittest.TestCase):
    """Test the initialization of the BrownianMotionFitter"""
    def setUp(self):
        self.reduced_f = f[::25000]
        self.reduced_PSD = PSDx[::25000]
        self.reduced_PSD_wgt = self.reduced_PSD * 0.25

    def test_BrownMotionFitter_init_success(self):
        # Try a case that should work.

        a = BrownianMotionFitter(self.reduced_f, self.reduced_PSD,
                                 self.reduced_PSD_wgt, T, est_cant)
        assert np.all(a.f == self.reduced_f)
        assert np.all(a.PSD_raw == self.reduced_PSD)
        assert np.all(a.T == T * u.K)
        assert a.est_cant == est_cant


ex_scaled_PSD = np.load('scale_data_PSD.npy')


class testBrownianMotionFitter_fitting(unittest.TestCase):
    def setUp(self):
        self.bmf = BrownianMotionFitter(f, PSDx, PSD_wgt, T, est_cant)

    def test_guess_P_detector(self):
        self.bmf._guess_P_detector()
        assert_almost_equal(self.bmf.P_detector0_raw, 6.5209499999999999e-09)

    def test_scale_data(self):
        self.bmf.P_detector0_raw = 6.5209499999999999e-09
        self.bmf._scale_data()
        np.testing.assert_allclose(ex_scaled_PSD, self.bmf.PSD)
        assert self.bmf.P_detector0 == 1

    def test_calc_initial_params(self):
        self.bmf.P_detector0_raw = 6.5209499999999999e-09
        self.bmf.P_detector0 = 1
        self.bmf.PSD = ex_scaled_PSD
        self.bmf._calc_initial_params()

        ex_initial_params = [8.917375e-05, 63700, 20000]

        np.testing.assert_allclose(ex_initial_params, self.bmf.initial_params)


def test_calc_k_c():
    """Test error calculation for calc_k_c.

    The calculation should use the (relative errors)

    P_x0: 0.01
    f_c: 1e-5
    Q: 0.01
    T: 0.01"""

    P_x0 = ufloat(1.75789868673e-12, 1.75789868673e-14) * u.nm**2/u.Hz  # 1/100
    f_c = ufloat(50000, 0.5) * u.Hz  # 1/100000 relative
    Q = ufloat(10000, 100) * u.dimensionless  # 1/100
    T = ufloat(300, 3) * u.K  # 1/100
    # ex_k_c is no longer a nice number because I switched from a rounded to
    # more exact value for Boltzmann's constant
    ex_k_c = ufloat(2.9999965233852217, 0.05196147267057527) * u.N/u.m
    k_c = calc_k_c(f_c, Q, P_x0, T)
    assert_almost_equal(k_c.magnitude.n, ex_k_c.magnitude.n)
    assert_almost_equal(k_c.magnitude.s, ex_k_c.magnitude.s)


def test_translate_fit_parameters():
    popt = np.array([1.75789868673e-6, 50e3, 10000, 1e-6])
    pcov = np.array([[1.75789868673e-8**2, 0, 0, 0],
                    [0, 0.5**2, 0, 0],
                    [0, 0, 100**2, 0],
                    [0, 0, 0, 1e-7**2]])
    P_detector0_raw = 1e-6
    T = ufloat(300, 3) * u.K
    ex_output = [ufloat(50000, 0.5) * u.Hz,
                 ufloat(2.9999965233852217, 0.05196147267057527) * u.N/u.m,
                 ufloat(10000, 100) * u.dimensionless,
                 ufloat(1e-12, 1e-13) * u.nm ** 2 / u.Hz]
    output = translate_fit_parameters(popt, pcov, P_detector0_raw, T)
    for ex_i, i in zip(ex_output, output):
        assert_almost_equal(ex_i.magnitude.n, i.magnitude.n)
        assert_almost_equal(ex_i.magnitude.s, i.magnitude.s)


def test_avg_ci_data():
    data = np.array([[1.1, 2.4, 3.95, 5.],
                     [1.2, 2.7, 3.9, 5.4],
                     [1.2, 2.4, 4., 5.7],
                     [1.3, 2.7, 3.8, 4.7]])
    ex_avg = np.array([1.2, 2.55, 3.9125, 5.2])
    ex_ci = np.array([0.08001666, 0.16974098, 0.08368343, 0.43090293])
    avg, ci = avg_ci_data(data)
    assert_array_almost_equal(ex_avg, avg)
    assert_array_almost_equal(ex_ci, ci)


def test_get_data():
    ex_f = np.array([1, 2, 3, 4, 5])
    ex_PSD = np.array([[1.1,  1.2, 1.2, 1.3],
                       [2.4,  2.7, 2.4, 2.7],
                       [3.95, 3.9, 4,   3.8],
                       [5,    5.4, 5.7, 4.7],
                       [5,    6,   6,   4]])

    # Create the test HDF5 file
    with h5py.File('test.h5', 'w', driver='core', backing_store=False) as fh:
        fh['x'] = ex_f
        fh['y'] = ex_PSD.mean(axis=1)
        fh['y'].attrs['n_avg'] = ex_PSD.shape[1]
        # Need ddof=1 to get sample standard deviation, rather than population.
        fh['y_std'] = ex_PSD.std(axis=1, ddof=1)

        ex_mean = np.array([1.2, 2.55, 3.9125, 5.2, 5.25])
        ex_ci = np.array([0.080016664930917122, 0.16974097914175013,
                          0.083683431255336824, 0.43090292797024871,
                          0.93827856560121137])

        f, psd_mean, psd_ci = get_data(fh)

    assert np.allclose(ex_f, f)
    assert np.allclose(ex_mean, psd_mean)
    assert np.allclose(ex_ci, psd_ci)




@unittest.skip("Currently support for the old data format is broken.")
class TestOldDataFormat(unittest.TestCase):
    """Test the old data format, with data stored in a group 'PSD', and
    individual power spectra stored underneath that as
    'PSD/d001', 'PSD/d002'..."""
    def setUp(self):
        # Filenames
        self.testv1 = 'test-data-v1.h5'  # Old data format
        self.testv2 = 'test-data-v2.h5'
        self.converted = 'converted-data.h5'
        # Sample data
        self.ex_f = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
        self.d1 = np.array([1.1, 2.4, 3.95, 5.,  5.])
        self.d2 = np.array([1.2, 2.7, 3.9, 5.4,  6.])
        self.d3 = np.array([1.2, 2.4, 4., 5.7,  6.])
        self.d4 = np.array([1.3, 2.7, 3.8, 4.7,  4.])

        self.ex_PSD = np.array([[1.1,  1.2, 1.2, 1.3],
                               [2.4,  2.7, 2.4, 2.7],
                               [3.95, 3.9, 4,   3.8],
                               [5,    5.4, 5.7, 4.7],
                               [5,    6,   6,   4]])
        # Create the HDF5 file test1
        with h5py.File(self.testv1, 'w') as fh:
            fh.create_group('PSD')
            fh.create_dataset('f', data=self.ex_f)
            fh.create_dataset('PSD/d001', data=self.d1)
            fh.create_dataset('PSD/d002', data=self.d2)
            fh.create_dataset('PSD/d003', data=self.d3)
            fh.create_dataset('PSD/d004', data=self.d4)

        # Create the test HDF5 file test2
        with h5py.File(self.testv2, 'w') as fh:
            fh['f'] = self.ex_f
            fh['PSD'] = self.ex_PSD

    def tearDown(self):
        silentremove(self.testv1)
        silentremove(self.testv2)
        silentremove(self.converted)

    def test_old_data_format_error(self):
        """Make sure that the function get_data throws a useful error when it
        encounters an file with a PSD group."""
        assert_raises(ValueError, get_data, self.testv1)

    def test_convert_data(self):
        convert_data(self.testv1, self.converted)
        f, psd_mean, psd_ci = get_data(self.converted)

        ex_mean = np.array([1.2, 2.55, 3.9125, 5.2, 5.25])
        ex_ci = np.array([0.080016664930917122, 0.16974097914175013,
                      0.083683431255336824, 0.43090292797024871,
                      0.93827856560121137])

        assert np.allclose(self.ex_f, f)
        assert np.allclose(ex_mean, psd_mean)
        assert np.allclose(ex_ci, psd_ci)

    # Test should be skipped for now
    def test_covert_data_new_format(self):
        assert_raises(ValueError, convert_data, self.testv2, self.converted)


def test_fit_residuals():
    sorted_normal = np.load('normal-residuals.npy')  # Load normal residuals
    loc, scale = fit_residuals(sorted_normal)
    assert_almost_equal(loc, 0, places=2)
    assert_almost_equal(scale, 1, places=2)



# -*- coding: utf-8 -*-
"""
Noise Fitting Brownian Motion Data

Ryan Dwyer

2013-11-22

This document will implement John Marohn's Brownian motion noise
fitting routine in Python. The class ``BrownianMotionFitter`` implements the
fitting routine. The rest of the module consists of helper functions for
specific portions of the fitting procedure and data input / output.

Dependencies:

- numpy
- scipy
- matplotlib
- pint
- uncertainties
- h5py

"""

from __future__ import division
import os
import errno

import numpy as np
import scipy
import scipy.stats
import six
sp = scipy
from scipy.optimize import curve_fit
import matplotlib as mpl
# Backend should be set correctly here.
from uncertainties import correlated_values, ufloat
import h5py
import pint

u = pint.UnitRegistry()

k_B = u.boltzmann_constant


class Cantilever(object):
    """Implement a simple unit cantilever, and require all inputs."""

    k_B = k_B

    def __init__(self, f_c, k_c, Q):
        """Initialize the cantilever."""
        self.f_c = f_c
        self.k_c = k_c
        self.Q = Q

    # Properties of the cantilver
    @property
    def omega_c(self):
        """Return the angular resonance frequency of the cantilever."""
        return self.f_c * 2 * pi

    @property
    def Gamma_i(self):
        """Return the cantilever's intrinsic dissipation."""
        return (self.k_c / (self.omega_c * self.Q))

    def F_min(self, T, bandwidth=1*u.Hz):
        """Return the thermally limited minimum detectable
        force (pN).

        The optional bandwidth parameter allows determining
        a miniumun force over a broader or narrower bandwidth
        than 1 Hz."""
        return ((4 * self.k_B * self.Gamma_i * T * bandwidth) ** 0.5)

    # Representations of the cantilever
    def __str__(self):
        """Write out the cantilever as its most important parameters:
        resonance frequency, spring constant, quality factor and
        intrinsic friction."""
        return "f_c = {self.f_c}, k_c = {self.k_c}, Q = {self.Q}\
Gamma_i = {self.Gamma_i}".format(self=self)

    def __repr__(self):
        """Return a representation of the cantilever. Rounds the cantilever
        to 9 digits, so eval(repr(cantilever)) is not necessarily equal to
        cantilever."""
        return "Cantilever(f_c = {self.f_c}, k_c = {self.k_c}, \
Q = {self.Q})".format(self=self)


class BrownianMotionFitter(object):
    """This object fits Brownian motion position fluctuation data to
    calculate the spring constant of the cantilever and the
    detector noise floor.
    """

    def __init__(self, f, PSD, PSD_ci, T, est_cant):
        """
        f
            frequency (Hz)

        PSD
            Power spectral density of position fluctuations (nm^2 / Hz)

        PSD_ci
            The confidence intervals corresponding to the PSD data (nm^2 / Hz)

        T
            temperature (K)

        est_cant
            A unitted cantilever with f_c, k_c and Q defined as best guesses.

        """
        self.f = f
        self.PSD_raw = PSD
        self.PSD_ci_raw = PSD_ci
        self.T = T * u.K
        self.est_cant = est_cant

        self.rcParams = {}

    def _guess_P_detector(self):
        """Guess a low, but reasonable number for the detector noise floor.
        This is used to scale the data for fitting."""
        try:
            self.mask
        except AttributeError:
            # Make a dummy mask of all trues if no mask has been set
            self.mask = np.isfinite(self.PSD_raw)

        self.P_detector0_raw = np.percentile(self.PSD_raw[self.mask], 25)

    def _scale_data(self):
        """Scale the raw data for fitting, so that floating point number errors
        are not an issue."""
        self.PSD = self.PSD_raw / self.P_detector0_raw
        self.P_detector0 = 1

    def calc_fit(self, f_min=None, f_max=None):
        """Fit the power spectrum data over the frequency range
        f_min to f_min, using a 3 pass approach.

        If no f_min and f_max are provided, it fits over the entire range
        of input frequencies."""

        # Set f_min, f_max
        if f_min is None:
            self.f_min = self.f.min()
        else:
            self.f_min = f_min
        if f_max is None:
            self.f_max = self.f.max()
        else:
            self.f_max = f_max

        # Store a mask to aid fitting and plotting
        self.mask = mask = make_mask(self.f, self.f_min, self.f_max)

        self._guess_P_detector()
        self._scale_data()

        f = self.fit_f = self.f[mask]
        PSD = self.fit_PSD = self.PSD[mask]
        PSD_ci = self.PSD_ci = self.PSD_ci_raw[mask]
        self.fit_PSD_raw = self.PSD_raw[mask]

        self._calc_initial_params()

        self._first_pass(f, PSD, PSD_ci)
        self._second_pass(f, PSD, PSD_ci)
        self._final_pass(f, PSD, PSD_ci)
        self._prepare_output()

    def _calc_initial_params(self):
        """Use the estimates to calculate initial parameters to use for
        the fitting functions."""
        f_c = self.est_cant.f_c.to(u.Hz)
        k_c = self.est_cant.k_c.to(u.N/u.m)
        Q = self.est_cant.Q.magnitude

        P_x0guess = (calc_P_x0(f_c, k_c, Q, self.T) /
                    (self.P_detector0_raw * u.nm ** 2 / u.Hz))

        self.initial_params = [P_x0guess.magnitude, f_c.magnitude, Q]

    def _first_pass(self, f, PSD, PSD_ci):
        """Fit the power spectrum data using initial guesses for the
        parameters and a fixed noise floor. Fitting outputs stored in
        self.popt1, self.pcov1.

        We use user input guesses for the three cantilever parameters,
        and fix the thermal noise floor to a low but reasonable guess.
        Then, we fit the remaining three parameters :math:`P_{x0}, f_c, Q`.
        """

        def Pf_fixed_P_detector(f, P_x0, f_c, Q):
            return Pf(f, P_x0, f_c, Q, self.P_detector0)

        self.popt1, self.pcov1 = curve_fit(Pf_fixed_P_detector,
                                           f, PSD,
                                           p0=self.initial_params,
                                           sigma=PSD_ci)

    def _second_pass(self, f, PSD, PSD_ci):
        """Fit the power spectrum holding the center frequency :math:`f_c`
        fixed. Fitting outputs stored in self.popt2, self.pcov2."""
        f_c = self.popt1[1]

        def Pf_fixed_f_c(f, P_x0, Q, P_detector):
            return Pf(f, P_x0, f_c, Q, P_detector)

        p0 = [self.popt1[0], self.popt1[2], self.P_detector0]

        self.popt2, self.pcov2 = curve_fit(Pf_fixed_f_c,
                                           f, PSD,
                                           p0=p0,
                                           sigma=PSD_ci)

    def _final_pass(self, f, PSD, PSD_ci):
        """Fit the power spectrum allowing all parameters to
        determined by the fitting algorithm. Fitting outputs stored to
        self.popt, self.pcov.

        The fit data and raw fit data is stored to self.PSD_fit,
        self.PSD_fit_raw."""
        popt1, popt2 = self.popt1, self.popt2
        f_c = popt1[1]
        p0 = [popt2[0], f_c, popt2[1], popt2[2]]
        self.popt, self.pcov = curve_fit(Pf,
                                         f, PSD,
                                         p0=p0,
                                         sigma=PSD_ci)
        self.PSD_fit = Pf(f, *self.popt)
        self.PSD_fit_raw = self.PSD_fit * self.P_detector0_raw

    def _prepare_output(self):
        f_c, k_c, Q, P_detector = translate_fit_parameters(
            self.popt, self.pcov, self.P_detector0_raw, self.T)
        self.residuals = (self.PSD_fit - self.PSD[self.mask])
        self.reduced_residuals = self.residuals / self.PSD_fit
        self.reduced_residuals_sorted = np.sort(self.reduced_residuals)
        self._fit_residuals()
        self.f_c, self.k_c, self.Q, self.P_detector = f_c, k_c, Q, P_detector

    def report(self):
        return u"""
    Input
    -----------------------------------------------
    Temperature           T: {self.T:P}

    Estimates
    -----------------------------------------------
    Spring constant     k_c: {self.est_cant.k_c:P}
    Resonance frequency f_c: {self.est_cant.f_c:P}
    Quality factor        Q: {self.est_cant.Q:P}
    
    Fitting
    -----------------------------------------------
    Fit frequency min f_min: {f_min:P}
    Fit frequency max f_max: {f_max:P}

    Results
    -----------------------------------------------
    Resonance frequency f_c: {self.f_c:.1uSP}
    Spring constant     k_c: {self.k_c:.1uSP}
    Quality Factor        Q: {self.Q:.1uSP}
    Detector Noise         : {self.P_detector:.1uSP}
            """.format(self=self, f_min=self.f_min*u.Hz, f_max=self.f_max*u.Hz)

    def rst_report(self):
        return u"""
Report::

    Input
    -----------------------------------------------
    Temperature           T: {self.T:P}

    Estimates
    -----------------------------------------------
    Spring constant     k_c: {self.est_cant.k_c:P}
    Resonance frequency f_c: {self.est_cant.f_c:P}
    Quality factor        Q: {self.est_cant.Q:P}
    
    Fitting
    -----------------------------------------------
    Fit frequency min f_min: {f_min:P}
    Fit frequency max f_max: {f_max:P}

    Results
    -----------------------------------------------
    Resonance frequency f_c: {self.f_c:.1uSP}
    Spring constant     k_c: {self.k_c:.1uSP}
    Quality Factor        Q: {self.Q:.1uSP}
    Detector Noise         : {self.P_detector:.1uSP}
            """.format(self=self, f_min=self.f_min*u.Hz, f_max=self.f_max*u.Hz)

    def print_output(self):
        print(self.report())

# These plotting functions could use a refactor
# Lots of duplicated code, logic fairly separate from the rest of the class
    def plot_fit(self, filename=None):
        """Plot the calculated fit."""
        f = self.fit_f

        mpl.rcParams.update(self.rcParams)
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        
        ax.semilogy(f, self.fit_PSD_raw, f, self.PSD_fit_raw)
        
        ax.set_xlim(self.f_min, self.f_max)
        ax.set_xlabel('Frequency [Hz]')
        ax.set_ylabel(u'PSD [nm²/Hz]')
        
        fig.gca()
        fig.canvas.draw()
        fig.tight_layout(pad=0.25)

        if filename is not None:
            fig.savefig(filename)

        self.fit_fig, self.fit_ax = fig, ax
        return fig, ax

    def plot_residuals(self, filename=None):
        """Plot the residuals of the calculated fit."""

        mpl.rcParams.update(self.rcParams)
        import matplotlib.pyplot as plt
        
        fig, ax = plt.subplots()
        
        ax.plot(self.fit_f, self.residuals)
        
        ax.set_xlabel('Frequency [Hz]')
        ax.set_ylabel(u'Residual [nm²/Hz]')
        
        fig.gca()
        fig.canvas.draw()
        fig.tight_layout(pad=0.25)

        if filename is not None:
            fig.savefig(filename)
        return fig, ax


    def plot_reduced_residuals(self, filename=None):
        mpl.rcParams.update(self.rcParams)
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots()

        ax.plot(self.fit_f, self.reduced_residuals)

        ax.set_xlabel('Frequency [Hz]')
        ax.set_ylabel('Normalized Residual')

        fig.gca()
        fig.canvas.draw()
        fig.tight_layout(pad=0.25)

        if filename is not None:
            fig.savefig(filename)
        return fig, ax

    def plot_cdf(self, filename=None):
        x = self.reduced_residuals_sorted
        size = x.size
        y = np.arange(1, 1 + size) / size
        popt = self.p_residuals

        mpl.rcParams.update(self.rcParams)
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots()

        ax.plot(x, y, 'bo', x, cdf(x, *popt), 'g-')
        ax.set_xlabel("Reduced Residual")
        ax.set_ylabel("CDF")

        fig.gca()
        fig.canvas.draw()
        fig.tight_layout(pad=0.25)

        if filename is not None:
            fig.savefig(filename)
        return fig, ax


    def _fit_residuals(self):
        self.p_residuals = fit_residuals(self.reduced_residuals_sorted)
        print("""
Residuals
-------------------------------------
        Mean: {self.p_residuals[0]:.2e}
   Std. dev.: {self.p_residuals[1]:.2e}""".format(self=self))


def Pf(f, P_x0, f_c, Q, P_detector):
    """The equation to calculate :math:`P_x(f)` from the cantilever
    parameters and detector noise floor.

    f
        Frequency, the independent variable.

    P_x0
        The zero frequency power spectral density of position fluctuations.

    f_c
        The cantilever resonance frequency.

    Q
        The quality factor of the cantilever.

    P_detector
        The detector noise floor.
        """
    return (P_x0 * f_c**4 /
            ((f**2 - f_c**2)**2 + f**2 * f_c**2 / Q**2)
            + P_detector)


def make_mask(f, f_min, f_max):
    """Return a mask of the array f. The mask can be used
    to give only elements of f between fmin and fmax.

    Use the resulting mask to block off elements of f outside
    of the specified bounds, as shown below.

    >>> f = np.arange(10)
    >>> mask = make_mask(f, 1.5, 5.5)
    >>> f[mask]
    array([2, 3, 4, 5])
    """
    mask_low = f < f_max
    mask_high = f > f_min
    return np.logical_and(mask_high, mask_low)


def translate_fit_parameters(popt, pcov, P_detector0_raw, T=300*u.K):
    """Take the fit parameters and covariances, and converts them to
    SI values and errors for f_c, k_c, Q.

    Also makes sure all values are positive."""
    pvals = correlated_values(popt, pcov)
    punits = [u.nm**2 / u.Hz, u.Hz, u.dimensionless, u.nm**2 / u.Hz]
    scales = [P_detector0_raw, 1, 1, P_detector0_raw]

    def create_measurement(uncert_val, unit, scale):
        return (uncert_val.n * unit).plus_minus(uncert_val.s) * scale

    P_x0, f_c, Q, P_detector = [np.abs(
                                    create_measurement(uncert_val, unit, scale)
                                ) for
                                uncert_val, unit, scale in
                                zip(pvals, punits, scales)]

    k_c = np.abs(calc_k_c(f_c, Q, P_x0, T))
    return f_c, k_c, Q, P_detector


def calc_k_c(f_c, Q, P_x0, T=(300, 1)*u.K):
    """Calculate the spring constant, returning a pint quantity containing
    a value with uncertainty, if uncertainty is used on the input (ufloat)."""
    return ((2 * k_B * T) / (np.pi * f_c * Q * P_x0)).to(u.N / u.m)


def calc_P_x0(f_c, Q, k_c, T):
    """Calculate the thermal noise floor, returning a pint quantity in
    units of nm^2 / Hz."""
    return ((2 * k_B * T) / (np.pi * f_c * Q * k_c)).to(u.nm ** 2 / u.Hz)


def iterate_fit(func, x, y, p0, sigma_i, wgt):
    popt = [[]]
    pcov = [[]]
    popt[0], pcov[0] = curve_fit(func, x, y, p0=p0, sigma=sigma_i)
    for i in xrange(10):
        new_sigma = wgt * func(x, *popt[-1])
        opt, cov = curve_fit(func, x, y, p0=popt[-1], sigma=new_sigma)
        popt.append(opt)
        pcov.append(cov)
    PSD_fit = func(x, *popt[-1])
    return popt[-1], pcov[-1], PSD_fit


def avg_ci_data(data, axis=0):
    """Average the data along the given dimension, and calculate the 95
    percent confidence interval. Return the avg, ci arrays."""
    data_avg = data.mean(axis=axis)
    data_std = data.std(axis=axis, ddof=1)
    data_ci = data_std / data.shape[axis]**0.5 * 1.96
    return data_avg, data_ci


def get_data(filename_or_fh):
    """Extract power spectrum data from an HDF5 file.
    Return f, PSD_mean, PSD_conf_int.

    The HDF file should have the following structure:

    x
        A table containing the frequencies of each point in the power
        spectrum.

    y
        Contains power spectra PSD_0(f) through PSD_(n-1)(f), arranged
            as follows, with each power spectrum occupying 1 column. 
        PSD_0(0 Hz) PSD_1(O Hz)
        PSD_0(1 Hz) PSD_1(1 Hz) ...

    y_std
        Contains standard deviations for the data points in y


    See http://h5labview.sourceforge.net and http://www.h5py.org for
    more information."""
    is_filename = isinstance(filename_or_fh, six.string_types)
    if is_filename:
        fh = h5py.File(filename_or_fh, 'r')
    else:
        fh = filename_or_fh

    # Put some checks about old-style files here.
    f = fh['x'].value
    PSD_mean = fh['y'].value
    PSD_ci = fh['y_std'].value * 1.96 / fh['y'].attrs['n_avg']**0.5

    if is_filename:
        fh.close()

    return f, PSD_mean, PSD_ci


def cdf(x, loc, scale):
    """The cumulative probability function of the normal distribution."""
    return sp.stats.norm.cdf(x, loc=loc, scale=scale)

def fit_residuals(sorted_residuals):
    """Return the result of fitting the residuals to the normal cdf.
    """
    size = sorted_residuals.size
    y = np.arange(1, 1 + size) / size

    popt, _ = curve_fit(cdf, sorted_residuals, y)

    return popt


def convert_data(oldfile, newfile, fileformat='v1'):
    """Converts data from the v1 format to the v2 format"""
    oldf = h5py.File(oldfile, 'r')
    if oldf.get('PSD', getclass=True) is not h5py.Group:
        oldf.close()
        raise ValueError("It appears you have inputted a new style file.\
            The file will not be converted.")
    else:
        newf = h5py.File(newfile, 'w')
        newf['f'] = oldf['f'][:]
        # Include any dataset that starts with a small d
        # This should catch the default d001, d0002, d003,
        data_list = [value for key, value in oldf['PSD'].items()
                     if key.startswith('d')]
        newf['PSD'] = np.empty((oldf['f'].value.size, len(data_list)))

        for i, val in enumerate(data_list):
            newf['PSD'][:, i] = val[:]
        oldf.close()
        newf.close()

def silentremove(filename):
    """If a file exists, delete it. Otherwise, return nothing.
       See http://stackoverflow.com/q/10840533/2823213"""
    try:
        os.remove(filename)
    except OSError as e:  # this would be "except OSError, e:" before Python 2.6
        if e.errno != errno.ENOENT:  # errno.ENOENT = no such file or directory
            raise  # re-raise exception if a different error occured

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



# -*- coding: utf-8 -*-
"""
calck
=====

This is a command line tool to easily fit brownian motion data and output
a report. The default usage of the command should be,

.. code: bash

    calck file.h5 66300 66500 -Q 30000 -k 3.5

Should defaults be saved into a .calckrc file?

Yes, I think so. Then, Q, k, f_c can be filled in from the config file
if necessary.
"""
import os
import os.path
import io
import zipfile
import base64

from bunch import Bunch
import click
import h5py
import docutils.core

import bs4

import brownian
from brownian import u, Cantilever, BrownianMotionFitter, get_data, silentremove

def file_extension(filename):
    """Return the file extension for a given filename. For example,

    Input               Output
    -------------       ---------
    data.csv            csv
    data.tar.gz         gz
    .vimrc              (empty string)"""

    return os.path.splitext(filename)[1][1:]

def img2uri(html_text):
    """Convert any relative reference img tags in html_input to inline data uri.
    Return the transformed html, in utf-8 format."""

    soup = bs4.BeautifulSoup(html_text, "lxml")

    image_tags = soup.find_all('img')

    for image_tag in image_tags:
        image_path = image_tag.attrs['src']
        if 'http' not in image_path:
            base64_text = base64.b64encode(open(image_path, 'rb').read())
            ext = file_extension(image_path)
            
            image_tag.attrs['src'] = (
                "data:image/{ext};base64,{base64_text}".format(
                    ext=ext, base64_text=base64_text)
                )

    return soup.prettify("utf-8")

# Cross-platform configuration file location
config_file = os.path.expanduser("~/foo.ini")

def calck(filename, f_min, f_max,
        quality_factor, spring_constant, resonance_frequency, temperature):
    """Base function for the command line interface. TODO: Add numpy style
    param strings."""
    est_cant = Cantilever(f_c=resonance_frequency*u.Hz,
                          k_c=spring_constant*u.N/u.m,
                          Q=quality_factor*u.dimensionless)

    data = get_data(filename)
    bmf = BrownianMotionFitter(*data, T=temperature, est_cant=est_cant)
    bmf.calc_fit(f_min, f_max)
    return bmf

"""CLI object psuedo-code:
class BrownianCLI(object):
    def __init__(self, filename, f_min, f_max, output,
        quality_factor, spring_constant, resonance_frequency, temperature):
        pass
        # boilerplate, plus, pick resonance frequency
    

    def __repr__(self):
        return cli_reproduce stuff

    def report(self):
        if output is None:
            self._text_report()
        elif output == 'html'
            self._html_report()
        elif output == 'zip'
            self._zip_report()

    def _text_report(self):
        pass

    def _html_report(self):
        pass

    def _zip_report(self):
        pass
"""

@click.command()
@click.argument('filename', type=click.Path(exists=True))
@click.argument('f_min', type=float)
@click.argument('f_max', type=float)
@click.option('--output', '-o', type=str, default=None)
@click.option('--quality-factor', '-Q', default=30000.0, help="Est. quality factor [unitless]")
@click.option('--spring-constant', '-k', default=3.5, help="Est. spring const. [N/m]")
@click.option('--temperature', '-T', default=298, help='Temperature (298) [K]')
@click.option('--resonance-frequency', '-f', default=None, type=float, help="Est. cant. freq. [Hz]")
def cli(filename, f_min, f_max, output,
        quality_factor, spring_constant, resonance_frequency, temperature):
    # This function needs to be broken up into component pieces.
    # For example, the reproduce python, reproduce cli bits can easily be their
    # own functions (and in fact, may make more sense as __repr__ methods of
    # classes)
    # The report generator should also be its own function.
    if resonance_frequency is None:
        resonance_frequency = (f_min + f_max) / 2.0


    file_path = os.path.abspath(filename)
    with h5py.File(filename, 'r') as f:
        date = f.attrs['date']
        time = f.attrs['time'].replace(':', '')

    if output is None:
        cli_reproduce = """calck "{file_path}" {f_min} {f_max} -Q {Q} -k {k} -f {f} -T {T}""".format(file_path=file_path, f_min=f_min, f_max=f_max, Q=quality_factor,
    k=spring_constant, f=resonance_frequency, T=temperature)
    else:
        cli_reproduce = """calck "{file_path}" {f_min} {f_max} -o {output} -Q {Q} -k {k} -f {f} -T {T}""".format(file_path=file_path, f_min=f_min, f_max=f_max, Q=quality_factor,
    k=spring_constant, f=resonance_frequency, T=temperature, output=output) 
    
    python_reproduce = """\
    import brownian._calck
    bmf = brownian._calck.calck("{file_path}",
        f_min={f_min}, f_max={f_max},
        quality_factor={Q},
        spring_constant={k},
        resonance_frequency={f},
        temperature={T})

    print(bmf.report())
""".format(file_path=file_path, f_min=f_min, f_max=f_max, Q=quality_factor,
    k=spring_constant, f=resonance_frequency, T=temperature)    

    bmf = calck(filename, f_min, f_max,
        quality_factor, spring_constant, resonance_frequency, temperature)


    if output is None:
        text = u"""\
Generated by brownian version {version}.

Reproduce this analysis with the following shell command:
    
{cli_reproduce}

or with the following python code:

{python_reproduce}

{report}
""".format(cli_reproduce=cli_reproduce,
    python_reproduce=python_reproduce, report = bmf.rst_report(),
    version=brownian.__version__)
    
        click.echo(text)
        return 0
    else:

        files = Bunch(fit="fit.png", reduced_residuals="reduced_residuals.png",
            cdf='cdf.png')

        bmf.plot_fit(files.fit)
        bmf.plot_reduced_residuals(files.reduced_residuals)
        bmf.plot_cdf(files.cdf)

        ReST = u"""\
Report
======


{report}

.. image:: {files.fit}

.. image:: {files.reduced_residuals}

.. image:: {files.cdf}

Reproduce this analysis with the following shell command:

.. code:: bash
    
    {cli_reproduce}

or with the following python code:

.. code:: python

{python_reproduce} 


Generated by ``brownian`` version ``{version}``.

""".format(files=files, cli_reproduce=cli_reproduce,
    python_reproduce=python_reproduce, report = bmf.rst_report(),
    version=brownian.__version__)
        
        ext = file_extension(output)
        if ext not in ('zip', 'html'):
            click.echo('output file extension must be zip or html, not {0}'.format(ext))
            return -1
        elif ext == 'html':
            image_dependent_html = docutils.core.publish_string(ReST, writer_name='html')
            self_contained_html = unicode(img2uri(image_dependent_html), 'utf8')
    
            # io module instead of built-in open because it allows specifying
            # encoding. See http://stackoverflow.com/a/22288895/2823213
            with io.open(output, 'w', encoding='utf8') as f:
                f.write(self_contained_html)

            for filename in files.values():
                silentremove(filename)
            return 0

        elif ext == 'zip':
            files.html = date+'-'+time+'.html'
            files.rst = date+'-'+time+'.rst'

            image_dependent_html = unicode(
                docutils.core.publish_string(ReST, writer_name='html'), 'utf8'
                )
    
            with io.open(files.html, 'w', encoding='utf8') as f:
                f.write(image_dependent_html)

            with io.open(files.rst, 'w', encoding='utf8') as f:
                f.write(ReST)

            
            with zipfile.ZipFile(output, 'w') as z:
                for f in files.values():
                    z.write(f)

            # Delete non-zipped version of the files
            for f in files.values():
                silentremove(f)

            return 0



import h5py
import numpy as np

import unittest
from nose.tools import ok_
from numpy.testing import assert_array_almost_equal

from brownian import silentremove
from brownian.hdf5 import generate_sample_file


class Test_generate_sample_file(unittest.TestCase):
    """Test constructing a sample hdf5 file, to verify the file
    is constructed correctly."""
    filename = 'test-pos-PSD-1.h5'
    x = np.array([0, 0.5, 1.0, 1.5])
    y = np.array([1, 2.1, 2.9, 4])
    y_std = np.array([0.2, 0.3, 0.4, 0.5])

    file_attrs = {'date': "2014-07-23",
                  'time': '13:00:00',
                  'version': 'pos-PSD-1.0',
                  'instrument': 'PSB-B19-AFM',
                  'help': '''Nice help about the data structure; \
omitted here for brevity.'''}

    x_attrs = {'name': 'Frequency',
               'unit': 'Hz',
               'label': 'Frequency [Hz]',
               'label_latex': r'$f \: [\mathrm{Hz}]',
               'help': 'Frequency array for PSD.'}

    y_attrs = {'name': 'PSD',
               'n_avg': 4,
               'units': 'nm^2/Hz',
               'label': 'PSD [nm^2/Hz]',
               'label_latex': r'P_{\delta x} \: [\mathrm{nm}^2/\mathrm{Hz}]',
               'help': 'Power spectral density of position fluctuations.'}

    def setUp(self):
        generate_sample_file(self.filename)
        self.fh = h5py.File(self.filename, 'r')

    def test_datasets_groups(self):
        fh = self.fh
        ok_(isinstance(fh['x'], h5py.Dataset), "x should be a dataset")
        ok_(isinstance(fh['y'], h5py.Dataset), "f should be a dataset")

        assert_array_almost_equal(fh['x'].value, self.x)
        assert_array_almost_equal(fh['y'].value, self.y)
        assert_array_almost_equal(fh['y_std'].value, self.y_std)

    def test_attributes(self):
        fh = self.fh
        assert fh.attrs.items() == self.file_attrs.items()
        assert fh['x'].attrs.items() == self.x_attrs.items()
        assert fh['y'].attrs.items() == self.y_attrs.items()

    def tearDown(self):
        self.fh.close()
        silentremove(self.filename)



import unittest
from nose.tools import assert_raises, assert_almost_equal
from numpy.testing import assert_array_almost_equal
import h5py

from brownian import bayes

class testBayes(unittest.TestCase):
    def setUp(self):
        self.fh = h5py.File('../ex/brownian173033.h5', 'r')
    def test_integration(self):
        """Integration test: data extraction, sampling, and creation of plotting object."""
        d = bayes.fh2data(self.fh, 70521, 70601, 3.5, 20000,
              sigma_Q=15000, Pdet=1e-8,
              sigma_Pdet=3e-8, sigma_kc=2.5)
        traces = bayes.sample_pymc3(d, samples=10, njobs=1)
        ppb = bayes.PlotPyMCBrownian(d, traces, 'test-data')

    def tearDown(self):
        self.fh.close()



"""
brownian.hdf5
2014-07-23
Ryan Dwyer

This contains code to implement the hdf5 file structure used for the data
we work up. We will generate the example file using h5py, and then avoid having
to version control the binary hdf5.

HDF5 Format for Brownian Motion Data
====================================

Here is a sample of the HDF5 file format.

```
/
attributes: {'date': '2014-07-22',
             'time': '13:00:00',
             'version': 'pos-PSD-1.0',
             'instrument': 'PSB-B19-AFM',
             'help': 'Nice help about the data structure; omitted here for brevity.'}
        f
        attributes: {'unit': 'Hz',
                     'help': 'Frequency array for PSD.'}
        [0, 0.5, 1, 1.5]

    PSD/
    attributes: {'n': 4,
                 'units': 'nm^2/Hz',
                 'help': 'Power spectral density of position fluctuations.'}
        mean
            [1, 2.1, 2.9, 4]
        stdev
            [0.2, 0.3, 0.4, 0.5]

```
"""

import h5py
import numpy as np


def generate_sample_file(filename):
    fh = h5py.File(filename, 'w')

    # Set file attributes
    file_attrs = {'date':"2014-07-23",
                  'time': '13:00:00',
                  'version': 'pos-PSD-1.0',
                  'instrument': 'PSB-B19-AFM',
                  'help': '''Nice help about the data structure; \
omitted here for brevity.'''}
    update_h5py_attrs(fh.attrs, file_attrs)

    # Set frequency information
    fh['x'] = np.array([0, 0.5, 1.0, 1.5])

    x_attrs = {'name': 'Frequency',
               'unit': 'Hz',
               'label': 'Frequency [Hz]',
               'label_latex': r'$f \: [\mathrm{Hz}]',
               'help': 'Frequency array for PSD.'}
    update_h5py_attrs(fh['x'].attrs, x_attrs)

    # Set PSD information
    fh['y'] = np.array([1, 2.1, 2.9, 4])
    y_attrs = {'name': 'PSD',
               'n_avg': 4,
               'units': 'nm^2/Hz',
               'label': 'PSD [nm^2/Hz]',
               'label_latex': r'P_{\delta x} \: [\mathrm{nm}^2/\mathrm{Hz}]',
               'help': 'Power spectral density of position fluctuations.'}

    update_h5py_attrs(fh['y'].attrs, y_attrs)

    fh['y_std'] = np.array([0.2, 0.3, 0.4, 0.5])

    fh.close()


def update_h5py_attrs(h5py_attrs, attrs):
    """Update the h5py attributes in h5py_attrs by adding the attributes in 
    the dictionary attrs.

    This will overwrite existing attributes.
    """
    for key, val in attrs.viewitems():
        h5py_attrs[key] = val



#!/usr/bin/env python
# Enables the package to be used in develop mode
# See http://pythonhosted.org/setuptools/setuptools.html#development-mode
# See https://github.com/scikit-learn/scikit-learn/issues/1016
import os
from setuptools import setup

# See https://github.com/warner/python-versioneer
import imp
fp, pathname, description = imp.find_module('versioneer')
try:
    versioneer = imp.load_module('versioneer', fp, pathname, description)
finally:
    if fp: fp.close()


versioneer.VCS = 'git'
versioneer.versionfile_source = 'brownian/_version.py'
versioneer.versionfile_build = 'brownian/_version.py'
versioneer.tag_prefix = '' # tags are like 1.2.0
versioneer.parentdir_prefix = 'brownian-' # dirname like 'myproject-1.2.0'

def get_packages(package):
    """
    Return root package and all sub-packages.
    See https://github.com/un33k/django-uuslug/blob/master/setup.py.
    Using this to auotmatically find all packages.

    """
    return [dirpath
            for dirpath, dirnames, filenames in os.walk(package)
            if os.path.exists(os.path.join(dirpath, '__init__.py'))]

setup(name='brownian',
      description="",
      url='https://github.com/ryanpdwyer/brownian',
      version=versioneer.get_version(),
      author='Ryan Dwyer',
      author_email='ryanpdwyer@gmail.com',
      packages=get_packages('brownian'),
      install_requires=[
      'numpy', 'scipy', 'matplotlib', 'pint',
      'nose', 'h5py', 'uncertainties', 'click', 'BeautifulSoup4', 'docutils',
      'bunch', 'six'
      ],
      tests_require=['nose'],
      test_suite='nose.collector',
      license = 'MIT',
      include_package_data=True,
      entry_points="""
        [console_scripts]
        calck=brownian._calck:cli
        bayesk=brownian.bayes:pymc_brownian_cli
      """,
      cmdclass=versioneer.get_cmdclass(),
      )



# -*- coding: utf-8 -*-
"""Tests for the _calck.py module"""
import unittest
from nose.tools import eq_
from brownian._calck import img2uri, calck, cli, file_extension

def test_file_extension():
    filename_exp_output = {
        'data.csv': 'csv',
        'data.tar.gz': 'gz',
        '.vimrc': '',
        'Makefile': ''}

    for filename, exp_output in filename_exp_output.viewitems():
        eq_(file_extension(filename), exp_output)


class TestImg2Uri(unittest.TestCase):
    pass



# This file was generated by 'versioneer.py' from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

version_version = '0.3.3'
version_full = '26fe50ecda001fc3ea59ea03fd7d0274095e88b6'
def get_versions(default={}, verbose=False):
    return {'version': version_version, 'full': version_full}




# -*- coding: utf-8 -*-
from __future__ import division, absolute_import, print_function
import os
import io

import datetime
import copy
import platform
import matplotlib.pyplot as plt
from scipy.stats import norm
from scipy import stats
import docutils.core
import numpy as np
import click
import pandas as pd
import seaborn as sns
from collections import OrderedDict
import h5py
import pymc3 as pm
import pymc3.stats as pmstats
from brownian import u, calc_P_x0, Pf
sns.set_style("white")
from matplotlib.offsetbox import AnchoredText
from six.moves import cPickle as pickle
from statsmodels.nonparametric.smoothers_lowess import lowess
from six import string_types
from brownian._calck import img2uri

windows = 'Windows' == platform.system()

directory = os.path.split(__file__)[0]
model_code_dict_fname = os.path.join(directory, 'stan_model_code.pkl')

def gamma_mustd(mu, std):
    """Define the Gamma distribution in terms of its mean ``mu`` and standard deviation ``std``,
    rather than shape ``k``, and scale ``theta`.
    
    See https://en.wikipedia.org/wiki/Gamma_distribution"""
    k = (mu/std)**2
    theta = std**2/mu
    return stats.gamma(a=k, scale=theta)

def Pfi(d, samp, i):
    fc = d['mu_fc'] + samp['dfc'][i]
    Q = samp['Q'][i]
    kc = samp['kc'][i]
    return Pf(d['f'],
              calc_P_x0(fc*u.Hz, Q, kc*u('N/m'), d['T']*u.K).to('nm^2/Hz').magnitude,
              fc, Q, samp['Pdet'][i]*d['scale']
             )

def Pf_func_gamma(x, samp, d):
    """The equation to calculate :math:`P_x(f)` from the cantilever
    parameters and detector noise floor.

    f
        Frequency, the independent variable.

    P_x0
        The zero frequency power spectral density of position fluctuations.

    f_c
        The cantilever resonance frequency.

    Q
        The quality factor of the cantilever.

    P_detector
        The detector noise floor.
        """
    fc = d['mu_fc'] + samp['dfc']
    Q = samp['Q']
    kc = samp['kc']
    return Pf(x,
              calc_P_x0(fc*u.Hz, Q, kc * u('N/m'), d['T']*u.K).to('nm^2/Hz').magnitude,
              fc, Q, samp['Pdet'] * d['scale']
             )


def eval_samples(func, x, samples, data=None, **kwargs):
    """Evaluate the function ``func`` at x, for each sample."""
    N = len(samples.values()[0])
    M = len(x)
    z = np.zeros((N, M))
    for i in xrange(N):
        d = {key: val[i] for key, val in samples.items()}
        z[i, :] = func(x, d, data, **kwargs)
    return z


def fh2data(fh, fmin, fmax, kc, Q, Pdet=None, T=298,
            sigma_fc=5, sigma_kc=5, sigma_Q=10000, sigma_Pdet=None):
    f_all = fh['x'][:]
    m = (f_all > fmin) & (f_all < fmax)
    f = f_all[m]
    psd = fh['y'][:][m]
    M = int(round(fh['y'].attrs['n_avg']))

    N = f.size
    
    fc = f[np.argmax(psd)]

    # Scale data
    psd_scale = psd.mean()
    psd_scaled = psd / psd_scale
    
    if Pdet is None:
        mu_Pdet = np.percentile(psd_scaled, 25)
    else:
        mu_Pdet = Pdet / psd_scale


    if sigma_Pdet is None:
        sigma_Pdet = 5 * mu_Pdet
    else:
        sigma_Pdet = sigma_Pdet / psd_scale
        
    
    return {'fmin': fmin,
     'fmax': fmax,
     'N': N,
     'M': M,       
     'y': psd_scaled,
     'f': f,
     'scale': psd_scale,
     'mu_fc': fc,
     'mu_kc': kc,
     'mu_Q': Q,
     'mu_Pdet': mu_Pdet,  # scaled
     'sigma_fc': sigma_fc,
     'sigma_kc': sigma_kc,
     'sigma_Q': sigma_Q,
     'sigma_Pdet': sigma_Pdet,
     'T': T,
     }


def np2data(freq, PSD, N_averages, fmin, fmax, kc, Q, Pdet=None, T=298,
            sigma_fc=5, sigma_kc=5, sigma_Q=10000, sigma_Pdet=None):
    """
    Create input dictionary for bayesian curve fitting of power spectral density
    brownian motion data. Parameters explained below. Those labeled '(prior)'
    set the prior for the bayesian fitting.

    freq: Array of frequency data [Hz]
    PSD: Mean power spectral density data from N_averages
    N_averages: Number of averages
    fmin: Minimum frequency to fit [Hz]
    fmax: Maximum frequency to fit [Hz]
    kc: Estimated cantilever spring constant [N/m] (prior)
    Q: Estimated cantilever Q (prior)
    Pdet: Estimated detector noise floor. Estimated if not provided. (prior)
    T: Temperature [K] 
    sigma_fc: Standard deviation in estimated cantilever frequency (prior)
    sigma_Q: Standard deviation in estimated quality factor (prior)
    sigma_Pdet: Standard deviation in estimated detector noise [nm^2/Hz]. Estimated if not provided. (prior)
    """
    m = (freq > fmin) & (freq < fmax)
    f = freq[m]
    psd = PSD[m]

    M = int(round(N_averages)) # Force to integer
    N = f.size
    
    fc = f[np.argmax(psd)]

    # Scale data
    psd_scale = psd.mean()
    psd_scaled = psd / psd_scale
    
    if Pdet is None:
        mu_Pdet = np.percentile(psd_scaled, 25)
    else:
        mu_Pdet = Pdet / psd_scale


    if sigma_Pdet is None:
        sigma_Pdet = 5 * mu_Pdet
    else:
        sigma_Pdet = sigma_Pdet / psd_scale
        
    
    return {'fmin': fmin,
     'fmax': fmax,
     'N': N,
     'M': M,       
     'y': psd_scaled,
     'f': f,
     'scale': psd_scale,
     'mu_fc': fc,
     'mu_kc': kc,
     'mu_Q': Q,
     'mu_Pdet': mu_Pdet,  # scaled
     'sigma_fc': sigma_fc,
     'sigma_kc': sigma_kc,
     'sigma_Q': sigma_Q,
     'sigma_Pdet': sigma_Pdet,
     'T': T,
     }



def prior_func_pymc(xname, data, x):
    if xname == 'dfc':
        return norm.pdf(x, scale=data['sigma_fc'])
    elif xname == 'kc':
        return gamma_mustd(mu=data['mu_kc'], std=data['sigma_kc']).pdf(x)
    elif xname == 'Q':
        return gamma_mustd(mu=data['mu_Q'], std=data['sigma_Q']).pdf(x)
    elif xname == 'Pdet':
        return gamma_mustd(mu=data['mu_Pdet'], std=data['sigma_Pdet']).pdf(x)
    else:
        # Nothing to plot for logposterior
        return None

def range_from_pts(pts, percentile=1, pad=1.25):
    xmax = np.percentile(pts, 100-percentile)
    xmin = np.percentile(pts, percentile)
    xmid = (xmax + xmin) / 2.
    delta_x = (xmax - xmin) / 2.
    return xmid - delta_x * pad, xmid + delta_x * pad


def plot_all_traces(samp, data=None, prior_func=None):
    """
    Call prior_func(name, data, x) to get plot data.
    """
    N = len(samp)
    fig, axes = plt.subplots(nrows=N, ncols=N, figsize=(N*1.5, N*1.5))
    for i, (yname, y) in enumerate(samp.items()):
        ymin, ymax = range_from_pts(y)
        for j, (xname, x) in enumerate(samp.items()):
            xmin, xmax = range_from_pts(x)
            ax = axes[i][j]
            if i != j:
                if i < j:
                    ax.plot(x, y, '.', markersize=4, alpha=0.1)
                else:
                    sns.kdeplot(x, y, ax=ax)
                ax.set_yticklabels([''])
                ax.set_xticklabels([''])
                ax.set_xlim(xmin, xmax)
                ax.set_ylim(ymin, ymax)
                namey = AnchoredText(yname, 2, frameon=False)
                namex = AnchoredText(xname, 4, frameon=False)
                ax.add_artist(namey)
                ax.add_artist(namex)
            else:
                sns.kdeplot(x, ax=ax)
                if prior_func is not None and data is not None:
                    line = ax.get_lines()[0]
                    xdata = line.get_xdata()
                    ydata = prior_func(xname, data, xdata)
                    if ydata is not None:
                        ax.plot(xdata, ydata)
                ax.set_yticklabels([''])
                ax.set_xticklabels([''])
                name = AnchoredText(xname,2,  frameon=False)
                ax.add_artist(name)

            if j == 0:
                ax.set_ylabel(yname)

            if i == len(samp.items())-1:
                ax.set_xlabel(xname)

    fig.tight_layout()
    return fig, axes


def HDI_from_MCMC(posterior_samples, credible_mass):
    """ Compute highest density interval from a sample of representative values,
    estimated as the shortest credible interval.
    
    Takes Arguments posterior_samples (samples from posterior),
    and credible mass (normally .95). 
    """
    sorted_points = sorted(posterior_samples)
    ciIdxInc = np.ceil(credible_mass * len(sorted_points)).astype('int')
    nCIs = len(sorted_points) - ciIdxInc
    ciWidth = [0]*nCIs
    for i in range(nCIs):
        ciWidth[i] = sorted_points[i + ciIdxInc] - sorted_points[i]
        HDImin = sorted_points[ciWidth.index(min(ciWidth))]
        HDImax = sorted_points[ciWidth.index(min(ciWidth))+ciIdxInc]
    return (HDImin, HDImax)

def sample_pymc3(d, samples=2000, njobs=2):
    with pm.Model() as model:
        dfc = pm.Normal(mu=0.0, sd=d['sigma_fc'], name='dfc')
        Q = pm.Gamma(mu=d['mu_Q'], sd=d['sigma_Q'], name='Q')
        Pdet = pm.Gamma(mu=d['mu_Pdet'], sd=d['sigma_Pdet'], name='Pdet')
        kc = pm.Gamma(mu=d['mu_kc'], sd=d['sigma_kc'], name='kc')

        M = d['M']
        T = d['T']
        scale=d['scale']
        mu_fc = d['mu_fc']
        f = d['f']


        like = pm.Gamma(alpha=M, beta=(M/(((2 * 1.381e-5 * T) / (np.pi * Q * kc)) / scale * (dfc + mu_fc)**3 /
                    ((f * f - (dfc + mu_fc)**2) * (f * f - (dfc + mu_fc)**2) + f * f * (dfc + mu_fc)**2 / Q**2)
                    + Pdet)),
                            observed=d['y'],
                            name='like')

        start = pm.find_MAP()
        step = pm.NUTS(state=start)
        
        trace = pm.sample(samples, step=step, start=start, progressbar=True, njobs=njobs)
    return trace


class PlotPyMCBrownian(object):
    def __init__(self, d, traces, name):

        self.data = d
        self.traces = traces
        self.name = name
        
        self.samples = OrderedDict()

        varnames = [var for var in traces.varnames if '_log_' not in var]
        for var in varnames:
            self.samples[var] = traces[var]

        self.y = self.data['y'] * self.data['scale']

        self.f = self.data['f']


        self.y_samp = eval_samples(Pf_func_gamma, self.f,
                                   self.samples, self.data)

        self.y_percent = lambda p: np.percentile(self.y_samp, p, axis=0)

        self.y_mean = np.mean(self.y_samp, axis=0)

        self.reduced_residuals = (self.y - self.y_mean) / self.y_mean

    def plot(self, fname=None, bbox_inches='tight', **kwargs):
        f = self.f

        Pdet = self.samples['Pdet'].mean() * self.data['scale']

        fig, ax = plt.subplots()
        ax.semilogy(self.f, self.y, color='b')
        ax.semilogy(self.f, self.y_mean, color='g')
        ax.semilogy(self.f, Pdet + np.zeros_like(self.f), 'g--')
        ax.semilogy(self.f, self.y_mean - Pdet, 'g--')
        ax.set_xlabel("Frequency [Hz]")
        ax.set_ylabel(u"PSD [nm²/Hz]")
        if fname is not None:
            fig.tight_layout()
            fig.savefig(fname, bbox_inches=bbox_inches, **kwargs)

        return fig, ax

    def plot_traces(self, fname=None, bbox_inches='tight', **kwargs):
        # Should plot the MCMC traces.
        axes = pm.traceplot(self.traces)
        fig = axes[0, 0].get_figure()
        if fname is not None:
            fig.tight_layout()
            fig.savefig(fname, bbox_inches=bbox_inches, **kwargs)

        return fig, axes


    def plot_zoomed(self, zoom=10, fname=None, bbox_inches='tight', **kwargs):
        f = self.f

        Delta_f  = (self.data['mu_fc'] / self.samples['Q'].mean()) * zoom
        mu_f = self.data['mu_fc'] + self.samples['dfc'].mean()
        m = (f >= (mu_f - Delta_f)) & (f <= (mu_f + Delta_f))

        fig, ax = plt.subplots()
        ax.semilogy(self.f[m], self.y[m], color='b')
        ax.semilogy(self.f[m], self.y_mean[m], color='g')
        ax.fill_between(self.f[m], self.y_percent(2.5)[m],
                        self.y_percent(97.5)[m], color='g', alpha=0.5)
        ax.set_xlim(mu_f - Delta_f, mu_f + Delta_f)
        ax.set_xlabel("Frequency [Hz]")
        ax.set_ylabel(u"PSD [nm²/Hz]")
        if fname is not None:
            fig.tight_layout()
            fig.savefig(fname, bbox_inches=bbox_inches, **kwargs)

        return fig, ax

    def plot_residuals(self, Hz_fraction=15, fname=None, bbox_inches='tight',
                        **kwargs):
        f_total = np.max(self.f) - np.min(self.f)
        frac = min(Hz_fraction / f_total, 1)
        self.residuals_lowess = lowess(self.reduced_residuals,
                                       self.f,
                                       frac=frac,
                                       return_sorted=False)
        
        fig, ax = plt.subplots()
        ax.plot(self.f, self.reduced_residuals, 'b.', alpha=0.5)
        ax.plot(self.f, self.residuals_lowess, 'g-', linewidth=2)
        ax.set_xlabel("Frequency [Hz]")
        ax.set_ylabel("Reduced Residual")
        if fname is not None:
            fig.tight_layout()
            fig.savefig(fname, bbox_inches=bbox_inches, **kwargs)



    def plot_pairs(self, fname=None, bbox_inches='tight', **kwargs):
        fig, axes = plot_all_traces(self.samples, data=self.data,
                               prior_func=prior_func_pymc)

        if fname is not None:
            fig.tight_layout()
            fig.savefig(fname, bbox_inches=bbox_inches, **kwargs)

        return fig, axes



    def summary(self, start=0, ndigits=4):
        trace = self.traces
        roundto = ndigits
        batches=100
        alpha=0.05

        varnames = [name for name in trace.varnames if not name.endswith('_')]

        stat_summ = pmstats._StatSummary(roundto, batches, alpha)
        pq_summ = pmstats._PosteriorQuantileSummary(roundto, alpha)
        summ = []
        for var in varnames:
            # Extract sampled values
            sample = trace.get_values(var, burn=start, combine=True)

            summ.append('\n%s:\n\n' % var)

            summ.append(stat_summ.output(sample))
            summ.append(pq_summ.output(sample))

        return '\n'.join(summ)



    def report(self, outfile=None, outdir=None, clean=True):
        if outfile is None:
            outfile = self.name

        basename = os.path.splitext(outfile)[0]

        if outdir is not None and not os.path.exists(outdir):
            os.mkdir(outdir)

        if outdir is None:
            outdir=''

        html_fname = os.path.join(outdir, basename+'.html')

        fit_fname = os.path.join(outdir, basename+'-fit.png')
        fit_zoomed_fname = os.path.join(outdir, basename+'-fit_zoomed.png')
        residuals_fname = os.path.join(outdir, basename+'-resid.png')
        pairs_fname = os.path.join(outdir, basename+'-pairs.png')
        traces_fname = os.path.join(outdir, basename+'-traces.png')


        self.plot(fname=fit_fname)
        self.plot_zoomed(fname=fit_zoomed_fname)
        self.plot_residuals(fname=residuals_fname)
        self.plot_pairs(fname=pairs_fname)
        self.plot_traces(fname=traces_fname)

        indented_summary_str = ['    '+line for 
                                line in self.summary().split('\n')]



        body = """\
======================
Brownian motion report
======================

Summary
=======

::

    {summary}

::

    fc [Hz]: {fc:.2f}
    Pdet [nm²/Hz]: {Pdet:.2e}


Fit
---

.. image:: {fit_fname}

Fit zoomed
----------

.. image:: {fit_zoomed_fname}


Residuals
---------

.. image:: {residuals_fname}


Pair Plots
----------

.. image:: {pairs_fname}

Trace Plot
----------

.. image:: {traces_fname}


""".format(summary='\n'.join(indented_summary_str),
           Pdet=self.samples['Pdet'].mean()*self.data['scale'],
           fc=self.samples['dfc'].mean()+self.data['mu_fc'],
           fit_fname=fit_fname,
           fit_zoomed_fname=fit_zoomed_fname,
           residuals_fname=residuals_fname,
           pairs_fname=pairs_fname,
           traces_fname=traces_fname)

        
        image_dependent_html = docutils.core.publish_string(body, writer_name='html')
        self_contained_html = unicode(img2uri(image_dependent_html), 'utf8')

        with io.open(html_fname, 'w', encoding='utf8') as f:
            f.write(self_contained_html)

        if clean:
            for fname in [fit_fname, fit_zoomed_fname, 
                          residuals_fname, pairs_fname, traces_fname]:
            
                try:
                    os.remove(fname)
                except:
                    pass


# This looks pretty good. Let's add information on fh (dataset, etc),
# and write a cli script.

# Note: Full workup should be a simple combination of,
# 1. Fitting
# 2. Plot / fit report

@click.command(help="""\
Fit brownian motion data using Bayesian MCMC, using data from an HDF5 file.

Arguments:

\b
FILENAME            HDF5 datafile
FMIN                Min frequency to fit [Hz]
FMAX                Max frequency to fit [Hz]
""")
@click.argument('filename', type=click.Path(exists=True))
@click.argument('fmin', type=float)
@click.argument('fmax', type=float)
@click.option('--output', '-o', type=str, default=None)
@click.option('--temperature', '-T', default=298., help='Temperature (298) [K]')
@click.option('--spring-constant', '-k', default=3.5, help='Spring constant prior mean (3.5 N/m)')
@click.option('--quality-factor', '-Q', default=20000., help='Quality factor prior mean (20000)')
@click.option('--spring-constant-stdev', '-sk', default=5., help='Spring constant prior standard deviation (5 N/m)')
@click.option('--quality-factor-stdev', '-sQ', default=20000., help='Quality factor prior standard deviation (20000)')
@click.option('--resonance-frequency-stdev', '-sf', default=5., help='Resonance frequency prior standard deviation')
@click.option('--detector-noise', '-P', default=None, help='Detector noise prior mean (None, est. from data)')
@click.option('--detector-noise-stdev', '-sP', default=None, help='Detector noise prior standard deviation (None, est. from data)')
@click.option('--chains', '-c', default=2, help="MCMC chains to run (2)")
@click.option('--iterations', '-i', default=2000, help='MCMC iterations per chain (2000)')
@click.option('--clean/--no-clean', default=True, help='Clean intermediate png files.')
def pymc_brownian_cli(filename, fmin, fmax, output, spring_constant,
                          quality_factor, temperature, spring_constant_stdev,
                          quality_factor_stdev, resonance_frequency_stdev,
                          detector_noise, detector_noise_stdev,
                          chains, iterations, clean):
    kc = spring_constant
    Q = quality_factor
    T = temperature
    sigma_kc = spring_constant_stdev
    sigma_Q = quality_factor_stdev
    sigma_fc = resonance_frequency_stdev
    Pdet = detector_noise
    sigma_Pdet = detector_noise_stdev
    fh = h5py.File(filename, 'r')
    data = fh2data(fh, fmin, fmax, kc, Q, sigma_Q=sigma_Q, T=T, sigma_kc=sigma_kc,
                   sigma_fc=sigma_fc, Pdet=Pdet, sigma_Pdet=sigma_Pdet)

    traces = sample_pymc3(data, iterations, chains)

    name = os.path.splitext(filename)[0]

    pbb = PlotPyMCBrownian(data, traces, name)

    if output is None:
        output = os.path.splitext(filename)[0]+'.html'

    pbb.report(output, clean=clean)

